function checkHVML() {
    if (typeof(document.getElementByHVMLHandle) == "function") {
        if (typeof(HVML) == "object" && HVML.version >= 10) {
            let elemStatus = document.getElementByHVMLHandle('731128');
            let elemRunner = document.getElementByHVMLHandle('790715');
            if (elemStatus && elemRunner) {
                elemStatus.textContent = 'Ready';
                elemRunner.textContent = '@' + HVML.hostName + '/' +
                    HVML.appName + '/' + HVML.runnerName;
                return true;
            }
            else {
                console.log("A new HVML content loaded");
            }
        }
        else {
            console.log("Make sure to use the correct version of xGUI Pro");
        }
    }
    else {
        console.log("Make sure to use the tailored WebKit (2.34.1 or later) " +
                "released by HVML community and " +
                "the configuration option `ENABLE_HVML_ATTRS` is ON.");
    }

    return false;
}

if (checkHVML()) {
    HVML.onrequest = function (json) {
        const dom_update_ops = ['append', 'prepend', 'insertAfter',
              'insertBefore', 'displace'];

        msg = JSON.parse(json);
        console.log("HVML.onrequest operation: " + msg.operation);
        console.log("HVML.onrequest elementType: " + msg.elementType);

        if (msg.operation === 'load') {
            document.open();
            document.write(msg.data);
            document.close();

            console.log("checking elements with hvml-events attribute...");
            const interestedElements = document.querySelectorAll("[hvml-events]");
            if (interestedElements.length > 0)
                registerEventsListener(interestedElements);
            return { requestId: msg.requestId, state: "Ok" };
        }
        else if (msg.operation === 'writeBegin') {
            document.open();
            document.write(msg.data);
            return { requestId: msg.requestId, state: "Ok" };
        }
        else if (msg.operation === 'writeMore') {
            document.write(msg.data);
            return { requestId: msg.requestId, state: "Ok" };
        }
        else if (msg.operation === 'writeEnd') {
            document.write(msg.data);
            document.close();

            console.log("checking elements with hvml-events attribute...");
            const interestedElements = document.querySelectorAll("[hvml-events]");
            if (interestedElements.length > 0)
                registerEventsListener(interestedElements);
            return { requestId: msg.requestId, state: "Ok" };
        }
        else if (msg.operation === 'update') {
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (!updateProperty(elem, msg.property, msg.data))
                        return { requestId: msg.requestId, state: "BadRequest" };
                }
                else {
                    return { requestId: msg.requestId, state: "NotFound" };
                }

                return { requestId: msg.requestId, state: "Ok" };
            }
            else if (msg.elementType === 'handles') {
                let handles = msg.element.split(',');
                let nr_updated = 0;
                for (let i = 0; i < handles.length; i++) {
                    let elem = document.getElementByHVMLHandle(handles[i]);
                    if (elem) {
                        if (updateProperty(elem, msg.property, msg.data))
                            nr_updated++;
                    }
                }

                if (nr_updated == 0)
                    return { requestId: msg.requestId, state: "NotFound" };
                else if (nr_updated < handles.length)
                    return { requestId: msg.requestId, state: "PartialContent" };

                return { requestId: msg.requestId, state: "Ok" };
            }
        }
        else if (msg.operation === 'clear') {
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (!clearElement(elem, msg.property))
                        return { requestId: msg.requestId, state: "BadRequest" };
                }
                else {
                    return { requestId: msg.requestId, state: "NotFound" };
                }

                return { requestId: msg.requestId, state: "Ok" };
            }
            else if (msg.elementType === 'handles') {
                let handles = msg.element.split(',');
                let nr_cleared = 0;
                for (let i = 0; i < handles.length; i++) {
                    let elem = document.getElementByHVMLHandle(handles[i]);
                    if (elem) {
                        if (clearElement(elem, msg.property))
                            nr_cleared++;
                    }
                }

                if (nr_cleared == 0)
                    return { requestId: msg.requestId, state: "NotFound" };
                else if (nr_cleared < handles.length)
                    return { requestId: msg.requestId, state: "PartialContent" };

                return { requestId: msg.requestId, state: "Ok" };
            }
        }
        else if (msg.operation === 'erase') {
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (!eraseElement(elem, msg.property))
                        return { requestId: msg.requestId, state: "BadRequest" };
                }
                else {
                    return { requestId: msg.requestId, state: "NotFound" };
                }

                return { requestId: msg.requestId, state: "Ok" };
            }
            else if (msg.elementType === 'handles') {
                let handles = msg.element.split(',');
                let nr_erased = 0;
                for (let i = 0; i < handles.length; i++) {
                    let elem = document.getElementByHVMLHandle(handles[i]);
                    if (elem) {
                        if (eraseElement(elem, msg.property))
                            nr_erased++;
                    }
                }

                if (nr_erased == 0)
                    return { requestId: msg.requestId, state: "NotFound" };
                else if (nr_erased < handles.length)
                    return { requestId: msg.requestId, state: "PartialContent" };

                return { requestId: msg.requestId, state: "Ok" };
            }
        }
        else if (dom_update_ops.indexOf(msg.operation) !== -1) {
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (!updateDocument(elem, msg.operation, msg.data))
                        return { requestId: msg.requestId, state: "BadRequest" };
                }
                else {
                    return { requestId: msg.requestId, state: "NotFound" };
                }

                return { requestId: msg.requestId, state: "Ok" };
            }
        }
        else if (msg.operation === 'callMethod') {
            let data = null;
            let state = "Ok";
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (typeof(msg.data.method) != 'string')
                        state = "BadRequest";
                    else
                        data = callMethod(elem, msg.data.method, msg.data.arg);
                }
                else {
                    state = "NotFound";
                }

                return { requestId: msg.requestId, state: state, data: data };
            }
            else if (msg.elementType === 'void') {
                if (typeof(msg.data.method) != 'string')
                    state = "BadRequest";
                else
                    data = callMethod(document, msg.data.method, msg.data.arg);

                return { requestId: msg.requestId, state: state, data: data };
            }
        }
        else if (msg.operation === 'getProperty') {
            let data = null;
            let state = "Ok";
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    data = getProperty(elem, msg.property);
                }
                else {
                    state = "NotFound";
                }

                return { requestId: msg.requestId, state: state, data: data };
            }
            else if (msg.elementType === 'void') {
                data = getProperty(document, msg.property);

                return { requestId: msg.requestId, state: state, data: data };
            }
        }
        else if (msg.operation === 'setProperty') {
            let data = null;
            let state = "Ok";
            if (msg.elementType === 'handle' || msg.elementType === 'id') {
                let elem;
                if (msg.elementType === 'id')
                    elem = document.getElementById(msg.element);
                else
                    elem = document.getElementByHVMLHandle(msg.element);

                if (elem) {
                    if (typeof(msg.property) != 'string')
                        state = "BadRequest";
                    else
                        data = setProperty(elem, msg.property, msg.value);
                }
                else {
                    state = "NotFound";
                }

                return { requestId: msg.requestId, state: state, data: data };
            }
            else if (msg.elementType === 'void') {
                data = setProperty(document, msg.property, msg.data);
                return { requestId: msg.requestId, state: state, data: data };
            }
        }

        return { requestId: msg.requestId, state: "NotImplemented" };
    }

}

function registerEventsListener(elems)
{
    console.log("registerEventsListener called");

    elems.forEach (function (elem) {
        if (elem.hvmlHandle !== 0 || elem.id != "") {
            console.log("elem.hvmlEventList: " + elem.hvmlEventList.length);
            for (let i = 0; i < elem.hvmlEventList.length; i++) {
                let eventTokens = elem.hvmlEventList.item(i).split(':');
                let eventName = eventTokens[0];
                let eventOpts = eventTokens[1];

                var optList = [];
                if (eventOpts)
                    optList = eventOpts.split(',');

                elem.addEventListener(eventName, function (evt) {
                    let data = {
                        targetTagName: evt.target.tagName,
                        targetHandle: evt.target.hvmlHandleText,
                        targetId: evt.target.id,
                        targetClass: evt.target.className,
                        targetValue: evt.target.value,
                        timeStamp: evt.timeStamp,
                        details: evt};

                    // TODO: handle more options
                    if (optList.includes('prevent') && evt.cancelable)
                        evt.preventDefault();
                    if (optList.includes('stop'))
                        evt.stopPropagation();

                    if (evt.currentTarget.hvmlHandle == 0) {
                        HVML.post(evt.type, "id",
                                evt.currentTarget.id,
                                JSON.stringify(data));
                    }
                    else {
                        HVML.post(evt.type, "handle",
                                evt.currentTarget.hvmlHandleText,
                                JSON.stringify(data));
                    }
                });
            }
        }
        else {
            console.log("invalid hvmlHandle");
        }

    });
}

function updateProperty(elem, property, data)
{
    if (property === "textContent") {
        elem.textContent = data;
        return true;
    }
    else if (property.startsWith("attr.")) {
        let attr = property.substring(5);
        elem.setAttribute(attr, data);
        return true;
    }
    else if (property.startsWith("prop.")) {
        let name = property.substring(5);
        try {
            const method = new Function('ELEM', 'VALUE', 'ELEM.' + name + '=VALUE');
            method(elem, data);
        } catch (error) {
            console.error(error);
            return false;
        }

        return true;
    }

    return false;
}

function updateDocument(elem, op, data)
{
    let fragment = new DocumentFragment();
    let container = document.createElement("div");
    container.innerHTML = data;

    let nr_elements = container.children.length;
    if (nr_elements > 0) {

        const interestedElements = container.querySelectorAll(".hvml-events");
        if (interestedElements.length > 0)
            registerEventsListener(interestedElements);

        /* have child elements, discard any Text node out all children. */
        while (container.firstChild) {
            fragment.appendChild(container.firstChild);
        }

        switch (op) {
            case 'append':
                elem.appendChild(fragment);
                break;
            case 'prepend':
                elem.insertBefore(fragment, elem.firstChild);
                break;
            case 'insertBefore':
                elem.parentNode.insertBefore(fragment, elem);
                break;
            case 'insertAfter':
                elem.parentNode.insertBefore(fragment, elem.nextSibling)
                break;
            case 'displace':
                elem.replaceChildren();
                elem.appendChild(fragment);
                break;
            default:
                return false;
        }
    }
    else {
        /* no child elements, treat as Text node. */
        switch (op) {
            case 'append':
                elem.append(data);
                break;
            case 'prepend':
                elem.prepend(data);
                break;
            case 'insertBefore':
                elem.before(data);
                break;
            case 'insertAfter':
                elem.after(data);
                break;
            case 'displace':
                elem.replaceChildren();
                elem.append(data);
                break;
            default:
                return false;
        }
    }

    return true;
}

function clearElement(elem, property)
{
    if (property === "textContent") {
        elem.textContent = "";
    }
    else if (property.startsWith("attr.")) {
        const attr = property.substring(5);
        elem.setAttribute(attr, "");
    }
    else
        elem.replaceChildren();

    return true;
}

function eraseElement(elem, property)
{
    if (property === "textContent") {
        elem.textContent = "";
    }
    else if (property.startsWith("attr.")) {
        try {
            const attr = property.substring(5);
            elem.removeAttribute(attr);
        } catch (error) {
            console.error(error);
            return false;
        }
    }
    else
        elem.remove();

    return true;
}

function getProperty(obj, property)
{
    try {
        const func = new Function('OBJ', 'return OBJ.' + property);
        return func(obj);
    } catch (error) {
        console.error(error);
        return null;
    }
}

function setProperty(obj, property, value)
{
    try {
        console.log("about to set property `" + property + "` of an element with Id is " + obj.id + " to `" + value + "`");

        const func = new Function('OBJ', 'VALUE', 'OBJ.' + property + '=VALUE; return OBJ.' + property);
        return func(obj, value);
    } catch (error) {
        console.error(error);
        return null;
    }
}

function callMethod(elem, method, arg)
{
    let funcBody;

    if (method.startsWith("call:")) {
        funcBody = method.substring(5);
    }
    else if (arg) {
        funcBody = 'return ELEM.' + method + '(ARG)';
    }
    else {
        funcBody = 'return ELEM.' + method + '()';
    }

    try {
        const func = new Function('ELEM', 'ARG', funcBody);
        return func(elem, arg);
    } catch (error) {
        console.error(error);
        return null;
    }
}

